;
; VRetrace.asm          Vertical Retrace detection for Windows 95/98
;
; By Thiadmer Riemersma, for Windows Developer's Journal
;----------------------------------------------------------------------------

;                               I N C L U D E S
;----------------------------------------------------------------------------
.586p
        include vmm.inc
        include vwin32.inc
        include shell.inc
        include ..\include\vretrace.h   ; shared file for VxD, kernel driver
                                        ; and the application

;                       S T R U C T U R E S / CONSTANTS
;----------------------------------------------------------------------------

VRETRACE_MAJOR_VERSION  equ  1
VRETRACE_MINOR_VERSION  equ  1
VR_VERSION              =    (VRETRACE_MAJOR_VERSION shl 8) + VRETRACE_MINOR_VERSION


;                         P U B L I C   D A T A
;----------------------------------------------------------------------------

VXD_LOCKED_DATA_SEG

; System_Time           dd      0       ; System time value address

VXD_LOCKED_DATA_ENDS


;                  D E V I C E   D E C L A R A T I O N
;----------------------------------------------------------------------------
VXD_LOCKED_CODE_SEG

DECLARE_VIRTUAL_DEVICE VRETRACE,   \
        VRETRACE_MAJOR_VERSION,    \
        VRETRACE_MINOR_VERSION,    \
        VRetrace_Control,,         \
        UNDEFINED_INIT_ORDER



; VRETRACE_Control  - Device control procedure for the VxD.
;
; Exit: CF = error flag
;
; Destroys: EAX, EBX, ECX, EDX, ESI, EDI, Flags
;----------------------------------------------------------------------------
public VRetrace_Control
VRetrace_Control PROC NEAR
        Control_Dispatch SYS_DYNAMIC_DEVICE_INIT, VRETRACE_Device_Init
        Control_Dispatch SYS_DYNAMIC_DEVICE_EXIT, VRETRACE_Device_Exit
        Control_Dispatch W32_DEVICEIOCONTROL, VRETRACE_ioctl
        clc
        ret
VRetrace_Control ENDP

VXD_LOCKED_CODE_ENDS

;                        P A G E A B L E   C O D E
;----------------------------------------------------------------------------

VXD_PAGEABLE_CODE_SEG


; Test_VRetrace
;
; Entry: esi - DIOCParams block
;   The parameter is the number of vertical retraces to time. This time is
;   in microseconds, but it is accurate only to a millisecond.
;
; Exit: Carry - set indicates error (timeout on retrace)
;   The time period for the number of vertical retraces is stored. This
;   value is zero on error (timeout; video card does not support vertical
;   retraces).
;----------------------------------------------------------------------------
public Test_VRetrace
BeginProc Test_VRetrace
        push    eax
        push    ebx
        push    ecx
        push    edx
        push    esi

        ; set up a few registers
        VMMCall Get_System_Time
        mov     ebx, eax                ; EBX is start time (ms)
        mov     edx, 3dah               ; VGA input status register

        ; wait for the start of a retrace
    test_loop1:
        ; check timeout
        VMMCall Get_System_Time
        sub     eax, ebx                ; get time difference
        cmp     eax, 100                ; 100 ms have passed?
        jae     test_error              ; yes, quit with error
        ; check VRetrace
        in      al, dx
        test    al, 8
        jz      test_loop1

        ; wait for the end of the retrace
    test_loop2:
        ; check timeout
        VMMCall Get_System_Time
        sub     eax, ebx                ; get time difference
        cmp     eax, 100                ; 100 ms have passed?
        jae     test_error              ; yes, quit with error
        ; check VRetrace
        in      al, dx
        test    al, 8
        jnz     test_loop2

        ; retrace found, now get the  time between the retraces
        mov     eax, [esi].lpvInBuffer
        or      eax, eax                ; NULL handle passed?
        jz      test_notime             ; yes, quit
        mov     ecx, [eax]              ; get the value of the flag
        or      ecx, ecx                ; "wait-start" == 0?
        jz      test_notime             ; yes, quit

        ; get the current time mark
        VMMCall Get_System_Time         ; EAX = timestamp in ms
        push    eax                     ; save

        ; wait for the number of retraces in ECX
    test_wait_next:
        mov     edx, 3dah               ; VGA input status register
        ; wait for the end of the retrace
    test_wait1:
        in      al, dx
        test    al, 8
        jnz     test_wait1
        ; wait for the start of a retrace
    test_wait2:
        in      al, dx
        test    al, 8
        jz      test_wait2
        dec     ecx
        jnz     test_wait_next

        ; calculate the time interval
        VMMCall Get_System_Time         ; EAX = new timestamp in ms
        pop     ecx                     ; ECX = old timestamp
        sub     eax, ecx                ; EAX = time passed (in ms)
        imul    eax, 1000               ; convert to time in microseconds
        jmp     short test_quit         ; retrace found, exit

    test_notime:
        mov     eax, 1
        jmp     short test_quit         ; no error
    test_error:
        stc
        xor     eax, eax
        jmp     short test_exit
    test_quit:
        ; store EAX in lpvOutBuffer
        mov     ebx, [esi].lpvOutBuffer
        mov     [ebx], eax
        mov     ebx, [esi].lpcbBytesReturned
        mov     DWORD PTR [ebx], 4
        clc
    test_exit:
        pop     esi
        pop     edx
        pop     ecx
        pop     ebx
        pop     eax
        ret
EndProc Test_VRetrace


; Wait_VRetrace
;
; Entry: esi - DIOCParams block
;   The parameter is a flag that tells whether to wait for the start of a
;   a vertical retrace, the end of a retrace, or just somewhere inside the
;   vertical retrace.
;
; Exit: Carry - always clear
;----------------------------------------------------------------------------
public Wait_VRetrace
BeginProc Wait_VRetrace
        push    eax
        push    ebx
        push    ecx
        push    edx

        ; set up a few registers
        mov     edx, 3dah               ; VGA input status register

        ; get parameter
        mov     ebx, VRETRACE_WAIT_INSIDE ; preset
        mov     eax, [esi].lpvInBuffer
        or      eax, eax                ; NULL handle passed?
        jz      wait_start_retrace      ; yes, do not wait for a "next" retrace
        mov     ebx, [eax]              ; get the value of the flag

        ; 1. optionally wait for the end of the current retrace
        cmp     ebx, VRETRACE_WAIT_START ; must we (first) wait until the end?
        jne     wait_start_retrace      ; no, skip
    wait_loop1:
        in      al, dx
        test    al, 8                   ; retrace bit set?
        jnz     wait_loop1              ; yes, stay looping

        ; 2. wait until we're in a (new) vertical retrace
    wait_start_retrace:
    wait_loop2:
        in      al, dx
        test    al, 8
        jz      wait_loop2

        ; 3. wait until the end of the retrace just found
        cmp     ebx, VRETRACE_WAIT_END  ; must we wait until the start?
        jne     wait_quit               ; no, skip
    wait_loop3:
        in      al, dx
        test    al, 8
        jnz     wait_loop3

    wait_quit:
        clc
        pop     edx
        pop     ecx
        pop     ebx
        pop     eax
        ret
EndProc Wait_VRetrace


; VRETRACE_ioctl - Respond to IOcontrol messages sent by Win32 program.
;
; Entry: ecx = IOCTL code
;        esi -> DIOC block
;----------------------------------------------------------------------------
public VRETRACE_ioctl
BeginProc VRETRACE_ioctl

try_closehandle:
        cmp     ecx, IOCTL_CLOSE_HANDLE
        jne     try_openhandle
        ; Nothing to do for this one
        jmp     short VR_ioctl_success          ; exit successfully

try_openhandle:
        cmp     ecx, IOCTL_OPEN_HANDLE
        jne     try_getversion
        ; Nothing to do for this one
        jmp     short VR_ioctl_success          ; exit successfully

try_getversion:
        cmp     ecx, IOCTL_VRETRACE_VERSION
        jne     try_test_retrace
        ; store version in lpvOutBuffer
        mov     eax, [esi].lpvOutBuffer
        mov     DWORD PTR [eax], VR_VERSION
        mov     eax, [esi].lpcbBytesReturned
        mov     DWORD PTR [eax], 4
        jmp     short VR_ioctl_success          ; exit successfully

try_test_retrace:
        cmp     ecx, IOCTL_VRETRACE_TEST
        jne     try_wait_retrace
        call    Test_VRetrace
        jc      VR_ioctl_fail
        jmp     short VR_ioctl_success          ; exit successfully

try_wait_retrace:
        cmp     ecx, IOCTL_VRETRACE_WAIT
        jne     VR_ioctl_fail
        call    Wait_VRetrace
        jc      VR_ioctl_fail
        jmp     short VR_ioctl_success          ; exit successfully

VR_ioctl_success:                               ; return success, clear EAX
        xor     eax, eax                        ; return zero = success
        clc
        ret

VR_ioctl_fail:            ; return failure, EAX = error code
        mov     eax, 119  ; "The system does not support the command requested"
        stc
        ret
EndProc VRETRACE_ioctl


; VRETRACE_Device_Exit - Nothing to do (yet).
;----------------------------------------------------------------------------
public VRETRACE_Device_Exit
BeginProc VRETRACE_Device_Exit
        mov     eax, 1
        clc
        ret
EndProc VRETRACE_Device_Exit

VXD_PAGEABLE_CODE_ENDS


;          D E V I C E   I N I T I A L I Z A T I O N   C O D E
;----------------------------------------------------------------------------

VXD_ICODE_SEG

; VRETRACE_Device_Init - Initialization
;
; Entry: ebx -> System VM handle (not used)
;        edx -> Reference data from real mode init portion
;
; Exit: CF  = error flag (always clear)
;       EAX = return code (always 1)
;----------------------------------------------------------------------------
public VRETRACE_Device_Init
BeginProc VRETRACE_Device_Init
        ; Carry Flag indicates an error, so return with CF cleay
        mov     eax, 1
        clc
        ret
EndProc VRETRACE_Device_Init

VXD_ICODE_ENDS

        END
