/*
 * VRetrace.c  Vertical Retrace detection for Windows NT
 *
 * By Thiadmer Riemersma, for Windows Developer's Journal.
 */
#include <ntddk.h>
#include "..\include\vretrace.h"

#define DEVICE_NAME_STRING	L"VRetrace"
#define VR_VERSION          0x100

#define STATUS_FAILURE      0xc0000001Lu

int vr_Test(int Count)
{
static KTIMER Timer;
  LARGE_INTEGER timeout;
  __int64 microseconds = 0;
  ULONG interval;
  LARGE_INTEGER start, end;

  KeInitializeTimer(&Timer);
  timeout.QuadPart = (__int64) -1000000L;  /* increments of 100ns -> 100 ms */
  KeSetTimer(&Timer, timeout, NULL);

  /* check if vertical retraces are issued at all */
vretrace_test1:
  if (KeReadStateTimer(&Timer))
    return 0;                   /* timeout -> quit with error */

  __asm mov     edx, 3dah       // VGA input status register
  __asm in      al, dx
  __asm test    al, 8           // vertical retrace active?
  __asm jz      vretrace_test1  // no, wait until it is

vretrace_test2:
  if (KeReadStateTimer(&Timer))
    return 0;                   /* timeout -> quit with error */

  __asm mov     edx, 3dah       // VGA input status register
  __asm in      al, dx
  __asm test    al, 8           // vertical retrace active?
  __asm jnz     vretrace_test2  // yes, wait until it ends

  /* now count the requested number of retraces and determine the period */
  interval = KeQueryTimeIncrement();
  KeQueryTickCount(&start);
  while (Count-- > 0) {

    /* wait until the end of this retrace */
    __asm mov   edx, 3dah       // VGA input status register
vretrace_loop2:
    __asm in    al, dx
    __asm test  al, 8           // bit 3 set?
    __asm jnz   vretrace_loop2  // yes, continue waiting

    /* wait for the start of the next retrace */
    __asm mov   edx, 3dah       // VGA input status register
vretrace_loop3:
    __asm in    al, dx
    __asm test  al, 8           // bit 3 set?
    __asm jz    vretrace_loop3  // no, continue waiting
  } /* while */
  KeQueryTickCount(&end);
  microseconds = (__int64)end.QuadPart - (__int64)start.QuadPart;
  microseconds = (microseconds * (__int64)interval) / 10;

  if (microseconds == 0)
    microseconds = 1;           /* return 1 to say that vertical retraces are available */
  return (int)microseconds;
}

void vr_Wait(int WaitCode)
{
  if (WaitCode==VRETRACE_WAIT_START) {
    /* wait until the end of this retrace */
    __asm mov   edx, 3dah       // VGA input status register
vretrace_loop1:
    __asm in    al, dx
    __asm test  al, 8           // bit 3 set?
    __asm jnz   vretrace_loop1  // yes, continue waiting
  } /* if */

  __asm mov     edx, 3dah       // VGA input status register
vretrace_loop2:
  __asm in      al, dx
  __asm test    al, 8           // bit 3 set?
  __asm jz      vretrace_loop2  // no, continue waiting

  if (WaitCode==VRETRACE_WAIT_END) {
    /* wait until the end of this retrace */
    __asm mov   edx, 3dah       // VGA input status register
vretrace_loop3:
    __asm in    al, dx
    __asm test  al, 8           // bit 3 set?
    __asm jnz   vretrace_loop3  // yes, continue waiting
  } /* if */
}

/*
 * Main dispatch routine
 */
NTSTATUS VRetrace_Dispatch(IN  PDEVICE_OBJECT  DeviceObject,
                           IN  PIRP            pIrp)
{
  PIO_STACK_LOCATION pIrpStack;
  NTSTATUS Status;
                              // NOTE:  Use METHOD_BUFFERED ioctls.
  PULONG pIOBuffer;           // Pointer to transfer buffer
                              // (treated as an array of longs, unless otherwise case).
  ULONG InBufferSize;         // Amount of data avail. from caller.
  ULONG OutBufferSize;        // Max data that caller can accept.
  ULONG OutByteCount;         // Actual count of output bytes
  ULONG IoControlCode;        // Control code request to driver

  pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

  //  Set default return status
  Status       = STATUS_NOT_IMPLEMENTED;
  OutByteCount = 0;

  // Dispatch based on major function code.
  switch (pIrpStack->MajorFunction) {
    case IRP_MJ_CREATE:
      // no special Create processing
      Status = STATUS_SUCCESS;
      break;
    case IRP_MJ_CLOSE:
      // no special open/close processing
      Status = STATUS_SUCCESS;
      break;
    case IRP_MJ_DEVICE_CONTROL:

      // Size of buffer containing data from application
      InBufferSize  = pIrpStack->Parameters.DeviceIoControl.InputBufferLength;

      // Size of buffer for data to be sent to application
      OutBufferSize = pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;

      // NT copies inbuf here before entry and copies this to outbuf after
      // return, for METHOD_BUFFERED IOCTL's.
      pIOBuffer     = (PULONG)pIrp->AssociatedIrp.SystemBuffer;

      IoControlCode = pIrpStack->Parameters.DeviceIoControl.IoControlCode;
      switch (IoControlCode) {
        //--------------------------------------------------------
        // IOCTL command dispatch
        //--------------------------------------------------------
        case IOCTL_VRETRACE_VERSION:
          pIOBuffer[0] = VR_VERSION;
          OutByteCount = sizeof(ULONG);
          Status       = STATUS_SUCCESS;
          break;
        case IOCTL_VRETRACE_TEST:
          pIOBuffer[0] = vr_Test((int)pIOBuffer[0]);
          OutByteCount = sizeof(ULONG);
          Status       = STATUS_SUCCESS;
          break;
        case IOCTL_VRETRACE_WAIT:
          vr_Wait((int)pIOBuffer[0]);
          Status       = STATUS_SUCCESS;
          break;
        default:
          Status       = STATUS_INVALID_DEVICE_REQUEST;
          break;
      }
    break;
  }

  // We're done with I/O request. Record the status of the I/O action.
  pIrp->IoStatus.Status = Status;
  pIrp->IoStatus.Information = OutByteCount;

  // Don't boost priority when returning
  IoCompleteRequest(pIrp, IO_NO_INCREMENT);

  return Status;
}

void VRetrace_Unload(IN PDRIVER_OBJECT DriverObject)
{
  WCHAR DOSNameBuffer[] = L"\\DosDevices\\" DEVICE_NAME_STRING;
  UNICODE_STRING uniDOSString;

  RtlInitUnicodeString(&uniDOSString, DOSNameBuffer);
  IoDeleteSymbolicLink (&uniDOSString);
  IoDeleteDevice(DriverObject->DeviceObject);
}

/*
 * Driver Entry routine.
 *
 * This routine is called only once after the driver is initially
 * loaded into memory.  It allocates everything necessary for the
 * driver's operation.  In our case, it allocates memory for our IOPM
 * array. It also creates a symbolic link to the device driver.  This allows
 * a user mode application to access our driver using the \\.\VRetrace
 * notation.
 */
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject,
                     IN PUNICODE_STRING RegistryPath)
{
  PDEVICE_OBJECT deviceObject;
  NTSTATUS status;
  WCHAR NameBuffer[] = L"\\Device\\" DEVICE_NAME_STRING;
  WCHAR DOSNameBuffer[] = L"\\DosDevices\\" DEVICE_NAME_STRING;
  UNICODE_STRING uniNameString, uniDOSString;
  int n;

  //  Set up device driver name and device object.
  RtlInitUnicodeString(&uniNameString, NameBuffer);
  RtlInitUnicodeString(&uniDOSString, DOSNameBuffer);

  status = IoCreateDevice(DriverObject, 0, &uniNameString, VR_TYPE, 0, FALSE, &deviceObject);
  if (!NT_SUCCESS(status))
    return status;

  status = IoCreateSymbolicLink (&uniDOSString, &uniNameString);
  if (!NT_SUCCESS(status))
  	return status;

  //  Initialize the Driver Object with driver's entry points.
  DriverObject->MajorFunction[IRP_MJ_CREATE]         = VRetrace_Dispatch;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = VRetrace_Dispatch;
  DriverObject->DriverUnload                         = VRetrace_Unload;

  return STATUS_SUCCESS;
}
