/*
 * VTest.c      Demo program for Vertical Retrace detection under Windows
 *
 * By Thiadmer Riemersma, for Windows Developer's Journal.
 */
#include <windows.h>
#include <stdio.h>
#include "include/vretrace.h"

HANDLE LoadDriver(LPSTR DriverName)
{
  OSVERSIONINFO VerInfo;
  char          DevicePath[64];

  // Check operating system
  VerInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  GetVersionEx(&VerInfo);
  if (VerInfo.dwPlatformId == VER_PLATFORM_WIN32_NT) {

    wsprintf(DevicePath, "\\\\.\\%s", (LPSTR)DriverName);
    return CreateFile(DevicePath,
                      GENERIC_READ | GENERIC_WRITE,
                      FILE_SHARE_READ | FILE_SHARE_WRITE,
                      0,                     // Default security
                      OPEN_EXISTING,
                      FILE_FLAG_OVERLAPPED,  // Perform asynchronous I/O
                      0);                    // No template

  } else {

    // Otherwise, assume Windows95/98
    wsprintf(DevicePath, "\\\\.\\%s.vxd", (LPSTR)DriverName);
    return CreateFile(DevicePath, 0,0,0,
                      CREATE_NEW,
                      FILE_FLAG_DELETE_ON_CLOSE | FILE_FLAG_OVERLAPPED, 0);
  }
}

int main(void)
{
  HANDLE VRetrace_Handle;
  DWORD dwCycles;
  DWORD dwRetCount;
  DWORD dwReturned;
  DWORD dwWaitCode;
  DWORD dwVersion;
  DWORD Interval;
  BOOL success;
  HANDLE hProcess, hThread;
  DWORD ClassPriority;
  int ThreadPriority;
  DWORD starttime, time;
  int count;

  VRetrace_Handle = LoadDriver(VRETRACE_NAME);
  if (VRetrace_Handle == INVALID_HANDLE_VALUE) {
    printf("Cannot load the vertical retrace driver\n");
    return 1;
  }

  /* check the version (should be non-zero) */
  dwVersion = 0;
  success = DeviceIoControl(VRetrace_Handle, IOCTL_VRETRACE_VERSION,
                            NULL, 0,
                            &dwVersion, sizeof dwVersion,
                            &dwReturned, NULL);
  if (!success || dwRetCount==0 || dwVersion<0x100) {
    CloseHandle(VRetrace_Handle);
    printf("Invalid version (%lx) returned\n", dwVersion);
    return 1;
  }

  /* if we got the vertical retrace driver, check whether the video card
   * supports vertical retraces at all
   */
  dwCycles = 0;
  success = DeviceIoControl(VRetrace_Handle, IOCTL_VRETRACE_TEST,
                            &dwCycles, sizeof dwCycles,
                            &dwRetCount, sizeof dwRetCount,
                            &dwReturned, NULL);
  if (!success || dwRetCount==0) {
    CloseHandle(VRetrace_Handle);
    printf("No vertical retrace detected\n");
    return 1;
  }

  /* now wait for one second and count the number of retraces */
  dwWaitCode = VRETRACE_WAIT_START;     /* wait for *next* retrace
                                         * if one is already active */

  timeBeginPeriod(1);                   /* timeGetTime() is more accurate on
                                         * NT if timeBeginPeriod(1) is set */
  count = 0;

  /* increase the thread priority */
  hProcess = GetCurrentProcess();
  hThread = GetCurrentThread();
  ClassPriority = GetPriorityClass(hProcess);
  ThreadPriority = GetThreadPriority(hThread);
  SetPriorityClass(hProcess, REALTIME_PRIORITY_CLASS);
  SetThreadPriority(hThread, THREAD_PRIORITY_TIME_CRITICAL);

  DeviceIoControl(VRetrace_Handle, IOCTL_VRETRACE_WAIT,
                  &dwWaitCode, sizeof dwWaitCode, NULL, 0, &dwReturned, 0);
  Interval = 1000;              /* wait for 1 second */
  time = starttime = timeGetTime();
  while (time - starttime < Interval) {
    Sleep(5);       /* give up time slice immediately after a retrace */
    DeviceIoControl(VRetrace_Handle, IOCTL_VRETRACE_WAIT,
                    &dwWaitCode, sizeof dwWaitCode, NULL, 0, &dwReturned, 0);
    time = timeGetTime();
    count++;
  }
  Interval = (WORD)(time - starttime);

  /* reset the thread priority */
  SetPriorityClass(hProcess, ClassPriority);
  SetThreadPriority(hThread, ThreadPriority);

  printf("Counted %d retraces in %5.3f seconds\n", count, Interval/1000.0);
  return 0;
}

